/*
 * getdb.c  -  Read system definitions from database file
 *
 * Copyright (C) 1998-2001 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: getdb.c,v 1.2 2001/08/27 08:40:15 gkminix Exp $
 */

#include <common.h>
#include <nblib.h>
#include "makerom.h"
#include "doconfig.h"
#include "protini.h"



/* Variables private to this module */
static struct bootdef bootd;		/* boot definition record */
static char *protini = NULL;		/* protocol.ini contents */




/*
 * Table containing option names for "ldoptsX" parameters.
 */
static struct {
	char *name;
	int  *optptr[MAXLOADERS];
} opttbl[] = {
	{ "useint18",	{&bootd.loaders[0].useint18,
			 &bootd.loaders[1].useint18,
			 &bootd.loaders[2].useint18}},
	{ "cardinst",	{&bootd.loaders[0].cardinst,
			 &bootd.loaders[1].cardinst,
			 &bootd.loaders[2].cardinst}},
	{ "bootask",	{&bootd.loaders[0].bootask,
			 &bootd.loaders[1].bootask,
			 &bootd.loaders[2].bootask}},
	{NULL,		{NULL, NULL, NULL}}
};


/*
 * Read bootrom loader options. This routine gets called everytime
 * a "ldoptsX" parameter is found in the database file.
 */
static char *readopts(name, arg)
char *name;
char *arg;
{
  char *tok, *cp;
  int i, num;

  /* Isolate loader number from parameter name */
  assert(strlen(name) == 7 && isdigit(name[6]));
  num = name[6] - '1';
  assert(num >= 0 && num < MAXLOADERS);

  /* Find each option in the list */
  tok = strtok(arg, ",");
  while (tok) {
	/* Skip leading blanks */
	while (*tok && (*tok == ' ' || *tok == '\t'))
		tok++;
	cp = tok;
	/* Find end of option name */
	while (*cp && (*cp != ' ' && *cp != '\t'))
		cp++;
	if (*cp) {
		/* Skip trailing blanks */
		*(cp++) = '\0';
		while (*cp && (*cp == ' ' || *cp == '\t'))
			cp++;
		if (!*cp)
			return("invalid characters following option name");
	}
	/* Find option name in list */
	for (i = 0; opttbl[i].name != NULL; i++)
		if (!strcmp(tok, opttbl[i].name))
			break;
	if (opttbl[i].name == NULL)
		return("invalid option");
	/* Set option to TRUE */
	*opttbl[i].optptr[num] = TRUE;
	/* Proceed with next option */
	tok = strtok(NULL, ",");
  }
  return(NULL);
}



/* Available network driver interface types */
static char *dtypes[] = {
  "packet driver", "ndis driver", "undi driver", NULL
};



/* Available bus types */
static char *btypes[] = {
  "ISA", "EISA", "MCA", "PCI", NULL
};



/* Available types for output file */
static char *ftypes[] = {
  "binary", "flash", "intel hex", "motorola hex", "tektronix hex", NULL
};



/* Parameters in each bootrom section of database file */
static struct paramdef dbparams[] = {
  /* General bootrom parameters */
  { "kernel",      par_string,	NULL,	{&bootd.kernelname}},
  { "bustype",     par_enum,	btypes,	{(char **)&bootd.bus_type}},
  { "pcivendor",   par_int,	NULL,	{(char **)&bootd.pci_vendid}},
  { "pcidevice",   par_int,	NULL,	{(char **)&bootd.pci_devid}},
  { "pnpdevid",    par_string,	NULL,	{&bootd.pnp_devid}},
  { "netdriver",   par_string,	NULL,	{&bootd.netdrv.name}},
  { "netdrvtype",  par_enum,	dtypes, {(char **)&bootd.netdrv.drivertype}},

#if MAXLOADERS < 3
#error Invalid number of loaders
#endif

  /* Parameters for defining loaders */
  { "loader1",     par_string,	NULL,	{&bootd.loaders[0].name}},
  { "loader2",     par_string,	NULL,	{&bootd.loaders[1].name}},
  { "loader3",     par_string,	NULL,	{&bootd.loaders[2].name}},
  { "outname1",    par_string,	NULL,	{&bootd.loaders[0].outname}},
  { "outname2",    par_string,	NULL,	{&bootd.loaders[1].outname}},
  { "outname3",    par_string,	NULL,	{&bootd.loaders[2].outname}},
  { "outsize1",    par_int,	NULL,	{(char **)&bootd.loaders[0].outsize}},
  { "outsize2",    par_int,	NULL,	{(char **)&bootd.loaders[1].outsize}},
  { "outsize3",    par_int,	NULL,	{(char **)&bootd.loaders[2].outsize}},
  { "outtype1",	   par_enum,	ftypes,	{(char **)&bootd.loaders[0].outtype}},
  { "outtype2",	   par_enum,	ftypes,	{(char **)&bootd.loaders[1].outtype}},
  { "outtype3",    par_enum,	ftypes,	{(char **)&bootd.loaders[2].outtype}},
  { "ldopts1",     par_proc,	NULL,	{(char **)&readopts}},
  { "ldopts2",     par_proc,	NULL,	{(char **)&readopts}},
  { "ldopts3",     par_proc,	NULL,	{(char **)&readopts}},

#if MAXPROGS < 8
#error Invalid number of drivers
#endif

  /* Parameters for packet driver interface */
  { "pktdriver",   par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[0]}},
  { "pdprog1",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[1]}},
  { "pdprog2",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[2]}},
  { "pdprog3",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[3]}},
  { "pdprog4",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[4]}},
  { "pdprog5",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[5]}},
  { "pdprog6",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[6]}},
  { "pdprog7",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.prognames[7]}},
  { "pktargs",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[0]}},
  { "pdargs1",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[1]}},
  { "pdargs2",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[2]}},
  { "pdargs3",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[3]}},
  { "pdargs4",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[4]}},
  { "pdargs5",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[5]}},
  { "pdargs6",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[6]}},
  { "pdargs7",     par_string,	NULL,	{&bootd.netdrv.driverdefs.pd.progargs[7]}},
  { "pktminsize",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[0]}},
  { "pdminsize1",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[1]}},
  { "pdminsize2",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[2]}},
  { "pdminsize3",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[3]}},
  { "pdminsize4",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[4]}},
  { "pdminsize5",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[5]}},
  { "pdminsize6",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[6]}},
  { "pdminsize7",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.minsizes[7]}},
  { "pktmaxsize",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[0]}},
  { "pdmaxsize1",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[1]}},
  { "pdmaxsize2",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[2]}},
  { "pdmaxsize3",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[3]}},
  { "pdmaxsize4",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[4]}},
  { "pdmaxsize5",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[5]}},
  { "pdmaxsize6",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[6]}},
  { "pdmaxsize7",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.pd.maxsizes[7]}},

  /* Parameters for NDIS interface */
  { "protini",     par_string,	NULL,	{&protini}},
  { "ndisdriver",  par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[0]}},
  { "ndprog1",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[1]}},
  { "ndprog2",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[2]}},
  { "ndprog3",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[3]}},
  { "ndprog4",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[4]}},
  { "ndprog5",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[5]}},
  { "ndprog6",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[6]}},
  { "ndprog7",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.prognames[7]}},
  { "ndargs1",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[1]}},
  { "ndargs2",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[2]}},
  { "ndargs3",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[3]}},
  { "ndargs4",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[4]}},
  { "ndargs5",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[5]}},
  { "ndargs6",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[6]}},
  { "ndargs7",     par_string,	NULL,	{&bootd.netdrv.driverdefs.ndis.progargs[7]}},
  { "ndisminsize", par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[0]}},
  { "ndminsize1",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[1]}},
  { "ndminsize2",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[2]}},
  { "ndminsize3",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[3]}},
  { "ndminsize4",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[4]}},
  { "ndminsize5",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[5]}},
  { "ndminsize6",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[6]}},
  { "ndminsize7",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.minsizes[7]}},
  { "ndismaxsize", par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[0]}},
  { "ndmaxsize1",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[1]}},
  { "ndmaxsize2",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[2]}},
  { "ndmaxsize3",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[3]}},
  { "ndmaxsize4",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[4]}},
  { "ndmaxsize5",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[5]}},
  { "ndmaxsize6",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[6]}},
  { "ndmaxsize7",  par_long,	NULL,	{(char **)&bootd.netdrv.driverdefs.ndis.maxsizes[7]}},

  /* Parameters for UNDI interface */
  { "undiprog",    par_string,	NULL,	{&bootd.netdrv.driverdefs.undi.name}},
  { NULL,          par_null,	NULL,	{NULL}}
};



/*
 * Read one entry from the database file
 */
struct bootdef *getdb(name)
char *name;
{
  struct sectdef sect;
  char *namebuf, *cp;
  int i, j;

  /* Generate section record */
  namebuf = (char *)nbmalloc(strlen(name) + 9);
  sprintf(namebuf, "%s:makerom", name);
  sect.name = namebuf;
  sect.params = dbparams;
  sect.startsect = NULL;
  sect.endsect = NULL;

  /* Get entry from database */
  memset(&bootd, 0, sizeof(bootd));
  for (i = 0; i < MAXPROGS; i++) {
	bootd.netdrv.driverdefs.pd.minsizes[i] = -1L;
	bootd.netdrv.driverdefs.pd.maxsizes[i] = -1L;
	bootd.netdrv.driverdefs.ndis.minsizes[i] = -1L;
	bootd.netdrv.driverdefs.ndis.maxsizes[i] = -1L;
  }
  readdb(&sect, dbname);

  /* Check for correct and missing values */
  copystr(&bootd.name, name);
  checkaccess(&bootd.kernelname, config.bindir);
  if (!bootd.kernelname) {
	prnerr1("no or invalid kernel file name given in section <%s>",
								namebuf);
	exit(EXIT_DB);
  }
  if (bootd.pnp_devid != NULL && strlen(bootd.pnp_devid) == 0) {
	free(bootd.pnp_devid);
	bootd.pnp_devid = NULL;
  }
  if (bootd.bus_type == BUSTYPE_NONE)
	bootd.bus_type = BUSTYPE_ISA;
  if (bootd.bus_type == BUSTYPE_PCI) {
	if (bootd.pci_vendid == 0 || bootd.pci_devid == 0) {
		prnerr1("missing PCI vendor/device ID in section <%s>", namebuf);
		exit(EXIT_DB);
	}
	if (bootd.pnp_devid != NULL) {
		prnerr1("PnP device ID for PCI network card in section <%s>",
								namebuf);
		exit(EXIT_DB);
	}
  } else if (bootd.pci_vendid != 0 || bootd.pci_devid != 0) {
	prnerr1("PCI vendor/device ID specified for non-PCI device in section <%s>",
								namebuf);
	exit(EXIT_DB);
  }
  if (bootd.pnp_devid != NULL) {
	if (strlen(bootd.pnp_devid) != PNPIDLEN ||
	    !isalpha(bootd.pnp_devid[0]) ||
	    !isalpha(bootd.pnp_devid[1]) ||
	    !isalpha(bootd.pnp_devid[2]) ||
	    !isxdigit(bootd.pnp_devid[3]) ||
	    !isxdigit(bootd.pnp_devid[4]) ||
	    !isxdigit(bootd.pnp_devid[5]) ||
	    !isxdigit(bootd.pnp_devid[6])) {
		prnerr1("invalid PnP device ID in section <%s>", namebuf);
		exit(EXIT_DB);
	}
  }
  checkaccess(&bootd.netdrv.name, config.bindir);
  if (!bootd.netdrv.name) {
	prnerr1("no or invalid network driver name given in section <%s>",
								namebuf);
	exit(EXIT_DB);
  }
  if (bootd.netdrv.drivertype == DRVTYPE_NONE) {
	prnerr1("missing driver type in section <%s>", namebuf);
	exit(EXIT_DB);
  }

  /* Reorder the loader list and check for correct values */
  bootd.loadernum = 0;
  for (i = 0, j = -1; i < MAXLOADERS; i++) {
	if (bootd.loaders[i].name != NULL) {
		checkaccess(&bootd.loaders[i].name, config.bindir);
		if (bootd.loaders[i].name == NULL) {
			prnerr2("loader %d invalid in section <%s>",
								i, namebuf);
			exit(EXIT_DB);
		}
	}
	if (bootd.loaders[i].name == NULL) {
		if (bootd.loaders[i].outname != NULL) {
			free(bootd.loaders[i].outname);
			bootd.loaders[i].outname = NULL;
		}
		if (j < 0)
			j = i;
	} else {
		if (bootd.loaders[i].outname == NULL) {
			prnerr2("no output file given for loader %d in section <%s>",
								i, namebuf);
			exit(EXIT_DB);
		}
		if (bootd.loaders[i].outtype == OUT_NONE)
			bootd.loaders[i].outtype = OUT_BINARY;
		bootd.loadernum++;
		if (j >= 0) {
			bootd.loaders[j] = bootd.loaders[i];
			bootd.loaders[i].name = NULL;
			i = j;
			j = -1;
		}
	}
  }
  if (!bootd.loadernum) {
	prnerr1("no bootrom loader specified in section <%s>", namebuf);
	exit(EXIT_DB);
  }

  /* Check for correct values for the different network driver types */
  if (bootd.netdrv.drivertype == DRVTYPE_PD) {
	struct pktdrvdef *pdp = &(bootd.netdrv.driverdefs.pd);

	checkaccess(&(pdp->prognames[0]), config.netdrvdir);
	if (pdp->prognames[0] == NULL) {
		prnerr1("no or invalid packet driver specified in section <%s>",
								namebuf);
		exit(EXIT_DB);
	}

	pdp->prognum = 1;
	for (i = 1, j = -1; i < MAXPROGS; i++) {
		if (pdp->prognames[i] != NULL) {
			checkaccess(&(pdp->prognames[i]), config.utilsdir);
			if (pdp->prognames[i] == NULL) {
				prnerr2("packet driver program %d invalid in section <%s>",
								i, namebuf);
				exit(EXIT_DB);
			}
		}
		if (pdp->prognames[i] == NULL) {
			if (pdp->progargs[i] != NULL) {
				free(pdp->progargs[i]);
				pdp->progargs[i] = NULL;
			}
			if (j < 0)
				j = i;
		} else if (j >= 0) {
			pdp->prognames[j] = pdp->prognames[i];
			pdp->progargs[j] = pdp->progargs[i];
			pdp->minsizes[j] = pdp->minsizes[i];
			pdp->maxsizes[j] = pdp->maxsizes[i];
			pdp->prognames[i] = NULL;
			pdp->progargs[i] = NULL;
			pdp->minsizes[i] = -1L;
			pdp->maxsizes[i] = -1L;
			pdp->prognum++;
			i = j;
			j = -1;
		}
	}

	for (i = 0; i < pdp->prognum; i++) {
		if (pdp->minsizes[i] < 0L)
			pdp->minsizes[i] = -1L;
		if (pdp->maxsizes[i] < 0L)
			pdp->maxsizes[i] = -1L;
		if (pdp->maxsizes[i] == 0L ||
		    (i == 0 && pdp->minsizes[i] == 0L) ||
		    (pdp->minsizes[i] > 0L && pdp->maxsizes[i] > 0L &&
		     pdp->minsizes[i] > pdp->maxsizes[i])) {
			prnerr2("invalid packet driver execution sizes for program %d in section <%s>",
								i, namebuf);
			exit(EXIT_DB);
		}
	}
  } else if (bootd.netdrv.drivertype == DRVTYPE_NDIS) {
	struct ndisdef *np = &(bootd.netdrv.driverdefs.ndis);

	checkaccess(&(np->prognames[0]), config.netdrvdir);
	if (np->prognames[0] == NULL) {
		prnerr1("no or invalid NDIS driver specified in section <%s>",
								namebuf);
		exit(EXIT_DB);
	}

	if (protini == NULL) {
		prnerr1("missing \'protini\' entry in section <%s>\n",
								namebuf);
		exit(EXIT_DB);
	}
	if (!strncmp(protini, "file:", 5)) {
		cp = readprotini(&protini[5]);
		parseprotini(cp, &(np->protocolini), &(np->protinisize));
		free(cp);
	} else
		parseprotini(protini, &(np->protocolini), &(np->protinisize));

	np->prognum = 1;
	for (i = 1, j = -1; i < MAXPROGS; i++) {
		if (np->prognames[i] != NULL) {
			checkaccess(&(np->prognames[i]), config.utilsdir);
			if (np->prognames[i] == NULL) {
				prnerr2("NDIS program %d invalid in section <%s>",
								i, namebuf);
				exit(EXIT_DB);
			}
		}
		if (np->prognames[i] == NULL) {
			if (np->progargs[i] != NULL) {
				free(np->progargs[i]);
				np->progargs[i] = NULL;
			}
			if (j < 0)
				j = i;
		} else if (j >= 0) {
			np->prognames[j] = np->prognames[i];
			np->progargs[j] = np->progargs[i];
			np->minsizes[j] = np->minsizes[i];
			np->maxsizes[j] = np->maxsizes[i];
			np->prognames[i] = NULL;
			np->progargs[i] = NULL;
			np->minsizes[i] = -1L;
			np->maxsizes[i] = -1L;
			np->prognum++;
			i = j;
			j = -1;
		}
	}

	for (i = 0; i < np->prognum; i++) {
		if (np->minsizes[i] < 0L)
			np->minsizes[i] = -1L;
		if (np->maxsizes[i] < 0L)
			np->maxsizes[i] = -1L;
		if (np->maxsizes[i] == 0L ||
		    (i == 0 && np->minsizes[i] == 0L) ||
		    (np->minsizes[i] > 0L && np->maxsizes[i] > 0L &&
		     np->minsizes[i] > np->maxsizes[i])) {
			prnerr2("invalid NDIS execution sizes for program %d in section <%s>",
								i, namebuf);
			exit(EXIT_DB);
		}
	}
  } else if (bootd.netdrv.drivertype == DRVTYPE_UNDI) {
	checkaccess(&bootd.netdrv.driverdefs.undi.name, config.netdrvdir);
	if (bootd.netdrv.driverdefs.undi.name == NULL) {
		prnerr1("missing or invalid UNDI driver name in section <%s>",
								namebuf);
		exit(EXIT_DB);
	}
  }
  free(namebuf);
  return(&bootd);
}

